"""
GeoVexModel.

This pytorch lightning module implements the GeoVeX hexagonal autoencoder model.

References:
    [1] https://openreview.net/forum?id=7bvWopYY1H
"""

import math
from typing import TYPE_CHECKING, Callable

from srai._optional import import_optional_dependencies
from srai.embedders import Model

if TYPE_CHECKING:  # pragma: no cover
    import torch
    from torch import nn

try:  # pragma: no cover
    import torch  # noqa: F811
    from torch import nn  # noqa: F811

except ImportError:
    from srai.embedders._pytorch_stubs import nn, torch


def get_radius(i: int, j: int) -> int:
    """
    Calculates the radius of a cube given its coordinates.

    Ref https://www.redblobgames.com/grids/hexagons/#distances-axial

    Args:
        i (int): The x-coordinate of the cube.
        j (int): The y-coordinate of the cube.

    Returns:
        int: The radius of the cube.
    """
    origin = (0, 0, 0)
    target = (i, j, -i - j)
    return cube_distance(origin, target)


def cube_distance(a: tuple[int, int, int], b: tuple[int, int, int]) -> int:
    """
    Calculates the maximum distance between two points in a cube.

    Args:
        a (Tuple[int, int, int]): The first point.
        b (Tuple[int, int, int]): The second point.
    """
    vec = cube_subtract(a, b)
    return max(abs(vec[0]), abs(vec[1]), abs(vec[2]))


def cube_subtract(a: tuple[int, int, int], b: tuple[int, int, int]) -> tuple[int, int, int]:
    """
    Subtracts the coordinates of two 3D points and returns the resulting tuple.

    Args:
        a (Tuple[int, int, int]): The first point to subtract from.
        b (Tuple[int, int, int]): The second point to subtract.

    Returns:
        Tuple[int, int, int]: The resulting tuple of subtracted coordinates.
    """
    return a[0] - b[0], a[1] - b[1], a[2] - b[2]


def get_shape(r: int) -> int:
    """
    Get the shape of the embedding.

    Args:
        r (int): The radius of the hexagonal region.

    Returns:
        int: The shape of the embedding.
    """
    return 2 * r + 2


def build_mask_funcs(R: int) -> tuple[Callable[[int, int], float], ...]:
    """
    Build the mask functions for the loss function. These functions depend on the radius of the
    hexagonal region. They weight the loss function to give more importance to the center of the
    region.

    Args:
        R (int): Radius of the hexagonal region.

    Returns:
        Tuple[callable, callable]: The mask functions.
    """  # noqa: D205

    def w_dist(i: int, j: int) -> float:
        """
        The Distance Weighting Kernel. Equation (6) in [1].

        Args:
            i (int): row index of the first point
            j (int): column index of the first point

        Returns:
            float: The weight of the loss function.
        """
        q = j - R
        r = i - R
        r = get_radius(q, r)
        return 1 / (1 + r) if r <= R else 0

    def w_num(i: int, j: int) -> float:
        """
        The Numerosity Weighting Kernel. Equation (6) in [1].

        The 6 is the number of hexagons in a ring,
        which is multiplied by the ring number to get the total number of hexagons in the ring.

        Args:
            i (int): row index of the first point
            j (int): column index of the first point

        Returns:
            float: The weight of the loss function.
        """
        # r represents the ring number, which can be found using the ij index
        q = j - R
        r = i - R
        r = get_radius(q, r)
        return 1 / (6 * r) if r <= R and r > 0 else 1 if r == 0 else 0

    return w_dist, w_num


class GeoVeXLoss(nn.Module):  # type: ignore
    """
    The loss function for the GeoVeX model.

    Defined in [1]. Equations (4) and (7).
    """

    def __init__(self, R: int):
        """
        Initialize the GeoVeXLoss.

        Args:
            R (int): The radius of the hexagonal region.
        """
        super().__init__()
        self.R = R

        # strip out the padding from y,
        # so that the loss function is only calculated on the valid regions
        self.M = get_shape(self.R) - 1

        # register the mask functions as buffers
        # so that they are saved with the model
        w_dist_func, w_num_func = build_mask_funcs(self.R)

        self.register_buffer(
            "_w_dist_matrix",
            torch.tensor(
                [[w_dist_func(i, j) for j in range(self.M)] for i in range(self.M)],
                dtype=torch.float32,
            ),
        )

        self.register_buffer(
            "_w_num_matrix",
            torch.tensor(
                [[w_num_func(i, j) for j in range(self.M)] for i in range(self.M)],
                dtype=torch.float32,
            ),
        )

    def forward(
        self, pi: "torch.Tensor", lambda_: "torch.Tensor", y: "torch.Tensor"
    ) -> "torch.Tensor":
        """
        Forward pass of the loss function.

        Args:
            pi (torch.Tensor): The predicted pi tensor.
            lambda_ (torch.Tensor): The predicted lambda tensor.
            y (torch.Tensor): The target tensor.

        Returns:
            float: The loss value.
        """
        # trim the padding from y, pi, and lambda_
        y = y[:, :, : self.M, : self.M]
        pi = pi[:, :, : self.M, : self.M]
        lambda_ = lambda_[:, :, : self.M, : self.M]

        I0 = (y == 0).float()
        I_greater_0 = (y > 0).float()

        # torch.exp(-1 * lambda_) instead of torch.exp(lambda_). the paper has a typo, I think...
        log_likelihood_0 = I0 * torch.log(pi + (1 - pi) * torch.exp(-1 * lambda_))
        log_likelihood_greater_0 = I_greater_0 * (
            torch.log(1 - pi)
            - lambda_
            + y * torch.log(lambda_)
            - torch.lgamma(y + 1)  # this is the ln(factorial(y))
        )

        log_likelihood = log_likelihood_0 + log_likelihood_greater_0
        loss: torch.Tensor = -torch.sum(
            log_likelihood * self._w_dist_matrix * self._w_num_matrix
        ) / (torch.sum(self._w_dist_matrix) * torch.sum(self._w_num_matrix))
        return loss


class HexagonalConv2d(nn.Module):  # type: ignore
    """Hexagonal Convolutional Layer."""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int = 3,
        stride: int = 2,
        padding: int = 0,
        bias: bool = True,
        groups: int = 1,
    ):
        """
        Initialize the HexagonalConv2d. This is a convolutional layer with a hexagonal kernel.

        Args:
            in_channels (int): The number of input channels.
            out_channels (int): The number of output channels.
            kernel_size (int, optional): The size of the kernel. Defaults to 3.
            stride (int, optional): The stride of the convolution. Defaults to 2.
            padding (int, optional): The padding of the convolution. Defaults to 0.
            bias (bool, optional): Whether to use bias. Defaults to True.
            groups (int, optional): The number of groups. Defaults to 1.
        """
        from torch import nn

        super().__init__()

        if kernel_size != 3:
            raise NotImplementedError("kernel_size must be 3. Hexagonal kernel is 3x3.")

        self.conv = nn.Conv2d(
            in_channels, out_channels, kernel_size, stride, padding, bias=bias, groups=groups
        )
        self.register_buffer("hexagonal_mask", self._create_hexagonal_mask())

    @staticmethod
    def _create_hexagonal_mask() -> "torch.Tensor":
        """Create the hexagonal mask."""
        return torch.tensor([[0, 1, 1], [1, 1, 1], [1, 1, 0]], dtype=torch.float32)

    def forward(self, x: "torch.Tensor") -> "torch.Tensor":
        """
        Forward pass of the HexagonalConv2d.

        Args:
            x (torch.Tensor): The input tensor.

        Returns:
            torch.Tensor: The output tensor.
        """
        self.conv.weight = nn.Parameter(self.conv.weight * self.hexagonal_mask)
        return self.conv(x)


class HexagonalConvTranspose2d(HexagonalConv2d):
    """Hexagonal Transpose Convolutional Layer."""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int = 3,
        stride: int = 2,
        padding: int = 0,
        output_padding: int = 0,
        bias: bool = True,
    ):
        """
        Initialize the HexagonalConvTranspose2d.

        This is a transpose convolutional layer with a hexagonal kernel.

        Args:
            in_channels (int): The number of input channels.
            out_channels (int): The number of output channels.
            kernel_size (int, optional): The size of the kernel. Defaults to 3.
            stride (int, optional): The stride of the convolution. Defaults to 2.
            padding (int, optional): The padding of the convolution. Defaults to 0.
            output_padding (int, optional): The output padding of the convolution. Defaults to 0.
            bias (bool, optional): Whether to use bias. Defaults to True.
        """
        from torch import nn

        super(HexagonalConv2d, self).__init__()

        self.conv = nn.ConvTranspose2d(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            output_padding,
            bias=bias,
        )
        self.register_buffer("hexagonal_mask", self._create_hexagonal_mask())


class Reshape(nn.Module):  # type: ignore
    """Reshape layer."""

    def __init__(self, shape: tuple[int, ...]):
        """
        Initialize the Reshape layer.

        Args:
            shape (Tuple[int, ...]): The shape of the output tensor.
        """
        super().__init__()
        self.shape = shape

    def forward(self, x: "torch.Tensor") -> "torch.Tensor":
        """
        Forward pass of the Reshape layer.

        Args:
            x (torch.Tensor): Input tensor.

        Returns:
            torch.Tensor: Reshaped tensor.
        """
        return x.view(self.shape)


class GeoVeXZIP(nn.Module):  # type: ignore
    """GeoVeX Zero-Inflated Poisson Layer."""

    def __init__(self, in_dim: int, m: int, out_dim: int):
        """
        Initialize the GeoVeXZIP layer.

        Args:
            in_dim (int): The input dimension.
            m (int): The height and width of the tensor
            r (int): The radius of the hexagonal region.
            out_dim (int): The output dimension.
        """
        from torch import nn

        super().__init__()
        self.in_dim = in_dim
        self.M = m
        self.out_dim = out_dim
        self.pi = nn.Linear(in_dim, out_dim)
        self.lambda_ = nn.Linear(in_dim, out_dim)

    def forward(self, x: "torch.Tensor") -> tuple["torch.Tensor", "torch.Tensor"]:
        """
        Forward pass of the GeoVeXZIP layer.

        Args:
            x (torch.Tensor): The input tensor.

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: The predicted pi and lambda tensors.
        """
        _x = x.view(-1, self.M, self.M, self.in_dim)
        pi = torch.sigmoid(self.pi(_x)).view(
            -1,
            self.out_dim,
            self.M,
            self.M,
        )
        lambda_ = torch.exp(self.lambda_(_x)).view(
            -1,
            self.out_dim,
            self.M,
            self.M,
        )
        # clamp pi to avoid nan's
        clamped_pi = torch.clamp(pi, 1e-6, 1 - 1e-6)
        return clamped_pi, lambda_


class GeoVexModel(Model):
    """
    GeoVeX Model.

    This class implements the GeoVeX model. It is based on a convolutional autoencoder with a Zero-
    Inflated Poisson layer. The model is described in [1]. It takes a 3d tensor as input (counts of
    features per region) and outputs dense embeddings. The 3d tensor consists of the target region
    at the center and radius R neighbors around it.
    """

    def __init__(
        self,
        k_dim: int,
        radius: int,
        conv_layers: int = 2,
        emb_size: int = 32,
        learning_rate: float = 1e-5,
        conv_layer_size: int = 256,
    ):
        """
        Initialize the GeoVeX model.

        Args:
            k_dim (int): the number of input channels
            radius (int): the radius of the hexagonal region
            conv_layers (int, optional): The number of convolutional layers. Defaults to 2.
            emb_size (int, optional): The dimension of the inner embedding. Defaults to 32.
            learning_rate (float, optional): The learning rate. Defaults to 1e-5.
            conv_layer_size (int, optional): The size of the initial convolutional layer.
        """
        if k_dim < conv_layer_size:
            raise ValueError(f"k_dim must be greater than {conv_layer_size}")

        if conv_layers < 2:
            raise ValueError("conv_layers must be greater than 1")

        if radius < 2:
            raise ValueError("R must be greater than 1")

        import_optional_dependencies(
            dependency_group="torch", modules=["torch", "pytorch_lightning"]
        )
        from torch import nn

        super().__init__()

        self.k_dim = k_dim

        self.R = radius
        self.lr = learning_rate
        self.emb_size = emb_size

        # input size is 2R + 2
        self.M = get_shape(self.R)

        # calculate the padding to preserve the input size
        #  equation for output size with stride is
        #  out_size = (in_size - kernel_size + padding + stride) / stride
        stride = 2
        kernel_size = 3
        padding = math.ceil(((stride - 1) * self.M - stride + kernel_size) / 2)

        # find the size of the linear layer
        # equation is the output size of the conv layers
        in_size = self.M
        ll_padding = 0 if self.R < 5 else 1
        for _ in range(conv_layers - 1):
            out_size = math.floor(
                (in_size - kernel_size + ll_padding + stride) / stride,
            )
            in_size = out_size

        conv_sizes = [conv_layer_size * 2**i for i in range(conv_layers)]
        self.encoder = nn.Sequential(
            nn.BatchNorm2d(self.k_dim),
            nn.ReLU(),
            # have to add padding to preserve the input size
            HexagonalConv2d(self.k_dim, conv_sizes[0], kernel_size=3, stride=2, padding=padding),
            nn.BatchNorm2d(conv_layer_size),
            nn.ReLU(),
            *(
                # second conv block
                nn.Sequential(
                    HexagonalConv2d(
                        conv_sizes[i - 1],
                        conv_sizes[i],
                        kernel_size=3,
                        stride=2,
                        padding=ll_padding,
                    ),
                    nn.BatchNorm2d(conv_sizes[i]),
                    nn.ReLU(),
                )
                for i in range(1, conv_layers)
            ),
            # flatten
            nn.Flatten(),
            nn.Linear(out_size**2 * conv_sizes[-1], self.emb_size),
        )

        self.decoder = nn.Sequential(
            nn.Linear(self.emb_size, out_size**2 * conv_sizes[-1]),
            # maintain the batch size, but reshape the rest
            Reshape((-1, conv_sizes[-1], out_size, out_size)),
            # decoder has conv transpose layers - 1,
            # as the reshape layer is the first "transpose layer"
            *(
                nn.Sequential(
                    HexagonalConvTranspose2d(
                        conv_sizes[-1 * (i + 1)],
                        conv_sizes[-1 * (i + 2)],
                        kernel_size=3,
                        stride=2,
                        output_padding=1,
                        padding=ll_padding,
                    ),
                    nn.BatchNorm2d(conv_sizes[-1 * (i + 2)]),
                    nn.ReLU(),
                )
                for i in range(conv_layers - 1)
            ),
            GeoVeXZIP(conv_sizes[0], self.M, self.k_dim),
        )

        self._loss = GeoVeXLoss(self.R)

    def forward(self, x: "torch.Tensor") -> tuple["torch.Tensor", "torch.Tensor"]:
        """
        Forward pass of the GeoVeX model.

        Args:
            x (torch.Tensor): The input tensor. The dimensions are
                (batch_size, k_dim, R * 2 + 1, R * 2 + 1).

        Returns:
            torch.Tensor: The output tensor.
        """
        res: tuple[torch.Tensor, torch.Tensor] = self.decoder(self.encoder(x))
        return res[0], res[1]

    def training_step(self, batch: list["torch.Tensor"], batch_idx: int) -> "torch.Tensor":
        # sourcery skip: class-extract-method
        """
        Perform a training step. This is called by PyTorch Lightning.

        One training step consists of a forward pass, a loss calculation, and a backward pass.

        Args:
            batch (List[torch.Tensor]): The batch of data.
            batch_idx (int): The index of the batch.

        Returns:
            torch.Tensor: The loss value.
        """
        loss = self._loss.forward(*self.forward(batch), batch)
        self.log("train_loss", loss, on_step=True, on_epoch=True)
        return loss

    def validation_step(self, batch: list["torch.Tensor"], batch_idx: int) -> "torch.Tensor":
        """
        Perform a validation step. This is called by PyTorch Lightning.

        Args:
            batch (List[torch.Tensor]): The batch of data.
            batch_idx (int): The index of the batch.

        Returns:
            torch.Tensor: The loss value.
        """
        loss = self._loss.forward(*self.forward(batch), batch)
        self.log("validation_loss", loss, on_step=True, on_epoch=True)
        return loss

    def configure_optimizers(self) -> list["torch.optim.Optimizer"]:
        """
        Configure the optimizers. This is called by PyTorch Lightning.

        Returns:
            List[torch.optim.Optimizer]: The optimizers.
        """
        opt: torch.optim.Optimizer = torch.optim.Adam(
            self.parameters(),
            lr=self.lr,
        )
        return [opt]



"""Utility download function."""

import warnings
from pathlib import Path

import requests
from tqdm import tqdm


def download_file(
    url: str, fname: str, chunk_size: int = 1024, force_download: bool = True
) -> None:
    """
    Download a file with progress bar.

    Args:
        url (str): URL to download.
        fname (str): File name.
        chunk_size (str): Chunk size.
        force_download (bool): Flag to force download even if file exists.

    Source: https://gist.github.com/yanqd0/c13ed29e29432e3cf3e7c38467f42f51
    """
    if Path(fname).exists() and not force_download:
        warnings.warn("File exists. Skipping download.", stacklevel=1)
        return

    Path(fname).parent.mkdir(parents=True, exist_ok=True)
    resp = requests.get(
        url,
        headers={"User-Agent": "SRAI Python package (https://github.com/kraina-ai/srai)"},
        stream=True,
    )
    resp.raise_for_status()
    total = int(resp.headers.get("content-length", 0))
    with (
        open(fname, "wb") as file,
        tqdm(
            desc=fname.split("/")[-1],
            total=total,
            unit="iB",
            unit_scale=True,
            unit_divisor=1024,
        ) as bar,
    ):
        for data in resp.iter_content(chunk_size=chunk_size):
            size = file.write(data)
            bar.update(size)



"""
Highway2Vec embedder.

This module contains the embedder from the `highway2vec` paper [1].

References:
    1. https://doi.org/10.1145/3557918.3565865
"""

import json
from pathlib import Path
from typing import Any, Optional, Union

import geopandas as gpd
import pandas as pd

from srai._optional import import_optional_dependencies
from srai.embedders import Embedder, ModelT
from srai.exceptions import ModelNotFitException

from .model import Highway2VecModel


class Highway2VecEmbedder(Embedder):
    """Highway2Vec Embedder."""

    def __init__(self, hidden_size: int = 64, embedding_size: int = 30) -> None:
        """
        Init Highway2Vec Embedder.

        Args:
            hidden_size (int, optional): Hidden size in encoder and decoder. Defaults to 64.
            embedding_size (int, optional): Embedding size. Defaults to 30.
        """
        import_optional_dependencies(
            dependency_group="torch", modules=["torch", "pytorch_lightning"]
        )

        self._model: Optional[Highway2VecModel] = None
        self._hidden_size = hidden_size
        self._embedding_size = embedding_size
        self._is_fitted = False

    def transform(
        self,
        regions_gdf: gpd.GeoDataFrame,
        features_gdf: gpd.GeoDataFrame,
        joint_gdf: gpd.GeoDataFrame,
    ) -> pd.DataFrame:  # pragma: no cover
        """
        Embed regions using features.

        Args:
            regions_gdf (gpd.GeoDataFrame): Region indexes and geometries.
            features_gdf (gpd.GeoDataFrame): Feature indexes, geometries and feature values.
            joint_gdf (gpd.GeoDataFrame): Joiner result with region-feature multi-index.

        Returns:
            pd.DataFrame: Embedding and geometry index for each region in regions_gdf.

        Raises:
            ValueError: If any of the gdfs index names is None.
            ValueError: If joint_gdf.index is not of type pd.MultiIndex or doesn't have 2 levels.
            ValueError: If index levels in gdfs don't overlap correctly.
        """
        import torch

        self._validate_indexes(regions_gdf, features_gdf, joint_gdf)
        self._check_is_fitted()
        features_df = self._remove_geometry_if_present(features_gdf)

        self._model.eval()  # type: ignore
        embeddings = self._model(torch.Tensor(features_df.values)).detach().numpy()  # type: ignore
        embeddings_df = pd.DataFrame(embeddings, index=features_df.index)
        embeddings_joint = joint_gdf.join(embeddings_df)
        embeddings_aggregated = embeddings_joint.groupby(level=[0]).mean()

        return embeddings_aggregated

    def fit(
        self,
        regions_gdf: gpd.GeoDataFrame,
        features_gdf: gpd.GeoDataFrame,
        joint_gdf: gpd.GeoDataFrame,
        trainer_kwargs: Optional[dict[str, Any]] = None,
        dataloader_kwargs: Optional[dict[str, Any]] = None,
    ) -> None:
        """
        Fit the model to the data.

        Args:
            regions_gdf (gpd.GeoDataFrame): Region indexes and geometries.
            features_gdf (gpd.GeoDataFrame): Feature indexes, geometries and feature values.
            joint_gdf (gpd.GeoDataFrame): Joiner result with region-feature multi-index.
            trainer_kwargs (Optional[Dict[str, Any]], optional): Trainer kwargs. Defaults to None.
            dataloader_kwargs (Optional[Dict[str, Any]], optional): Dataloader kwargs.
                Defaults to None.

        Raises:
            ValueError: If any of the gdfs index names is None.
            ValueError: If joint_gdf.index is not of type pd.MultiIndex or doesn't have 2 levels.
            ValueError: If index levels in gdfs don't overlap correctly.
        """
        import pytorch_lightning as pl
        import torch
        from torch.utils.data import DataLoader

        self._validate_indexes(regions_gdf, features_gdf, joint_gdf)
        features_df = self._remove_geometry_if_present(features_gdf)

        num_features = len(features_df.columns)
        self._model = Highway2VecModel(
            n_features=num_features, n_hidden=self._hidden_size, n_embed=self._embedding_size
        )

        dataloader_kwargs = dataloader_kwargs or {}
        if "batch_size" not in dataloader_kwargs:
            dataloader_kwargs["batch_size"] = 128

        dataloader = DataLoader(torch.Tensor(features_df.values), **dataloader_kwargs)

        trainer_kwargs = trainer_kwargs or {}
        if "max_epochs" not in trainer_kwargs:
            trainer_kwargs["max_epochs"] = 10

        trainer = pl.Trainer(**trainer_kwargs)
        trainer.fit(self._model, dataloader)
        self._is_fitted = True

    def fit_transform(
        self,
        regions_gdf: gpd.GeoDataFrame,
        features_gdf: gpd.GeoDataFrame,
        joint_gdf: gpd.GeoDataFrame,
        trainer_kwargs: Optional[dict[str, Any]] = None,
        dataloader_kwargs: Optional[dict[str, Any]] = None,
    ) -> pd.DataFrame:
        """
        Fit the model to the data and return the embeddings.

        Args:
            regions_gdf (gpd.GeoDataFrame): Region indexes and geometries.
            features_gdf (gpd.GeoDataFrame): Feature indexes, geometries and feature values.
            joint_gdf (gpd.GeoDataFrame): Joiner result with region-feature multi-index.
            trainer_kwargs (Optional[Dict[str, Any]], optional): Trainer kwargs. Defaults to None.
            dataloader_kwargs (Optional[Dict[str, Any]], optional): Dataloader kwargs.
                Defaults to None.

        Returns:
            pd.DataFrame: Region embeddings.

        Raises:
            ValueError: If any of the gdfs index names is None.
            ValueError: If joint_gdf.index is not of type pd.MultiIndex or doesn't have 2 levels.
            ValueError: If index levels in gdfs don't overlap correctly.
        """
        self.fit(regions_gdf, features_gdf, joint_gdf, trainer_kwargs, dataloader_kwargs)
        return self.transform(regions_gdf, features_gdf, joint_gdf)

    def _check_is_fitted(self) -> None:
        if not self._is_fitted or self._model is None:
            raise ModelNotFitException("Model not fitted. Call fit() or fit_transform() first.")

    def _save(self, path: Union[Path, str], embedder_config: dict[str, Any]) -> None:
        if isinstance(path, str):
            path = Path(path)

        self._check_is_fitted()

        path.mkdir(parents=True, exist_ok=True)

        self._model.save(path / "model.pt")  # type: ignore

        config = {
            "model_config": self._model.get_config(),  # type: ignore
            "embedder_config": embedder_config,
        }
        with (path / "config.json").open("w") as f:
            json.dump(config, f, ensure_ascii=False, indent=4)

    def save(self, path: Union[Path, str]) -> None:
        """
        Save the model to a directory.

        Args:
            path (Path): Path to the directory.
        """
        embedder_config = {"hidden_size": self._hidden_size, "embedding_size": self._embedding_size}
        self._save(path, embedder_config)

    @classmethod
    def _load(cls, path: Union[Path, str], model_module: type[ModelT]) -> "Highway2VecEmbedder":
        if isinstance(path, str):
            path = Path(path)

        with (path / "config.json").open("r") as f:
            config = json.load(f)
        embedder = cls(**config["embedder_config"])
        model_path = path / "model.pt"
        model = model_module.load(model_path, **config["model_config"])
        embedder._model = model
        embedder._is_fitted = True
        return embedder

    @classmethod
    def load(cls, path: Union[Path, str]) -> "Highway2VecEmbedder":
        """
        Load the model from a directory.

        Args:
            path (Path): Path to the directory.

        Returns:
            Hex2VecEmbedder: The loaded embedder.
        """
        return cls._load(path, Highway2VecModel)



"""Tests for AdjacencyNeighbourhood."""

import geopandas as gpd
import pytest
from shapely import geometry

from srai.constants import WGS84_CRS
from srai.neighbourhoods import AdjacencyNeighbourhood


@pytest.fixture  # type: ignore
def no_geometry_gdf() -> gpd.GeoDataFrame:
    """Get GeoDataFrame with no geometry."""
    return gpd.GeoDataFrame()


@pytest.fixture  # type: ignore
def empty_gdf() -> gpd.GeoDataFrame:
    """Get empty GeoDataFrame."""
    return gpd.GeoDataFrame(geometry=[])


@pytest.fixture  # type: ignore
def squares_regions_fixture() -> gpd.GeoDataFrame:
    """
    Get 9 square regions.

    This GeoDataFrame represents 9 squares on a cartesian plane in a 3 by 3 grid pattern. Squares
    are adjacent by edges and by vertices. Squares are given compass directions. Visually it looks
    like this ("C" means "CENTER"):

    [["NW", "N", "NE"],
     [ "W", "C",  "E"],
     ["SW", "S", "SE"]]

    Returns:
        GeoDataFrame: A GeoDataFrame containing square regions.
    """
    regions = gpd.GeoDataFrame(
        geometry=[
            geometry.box(minx=0, maxx=1, miny=0, maxy=1),
            geometry.box(minx=1, maxx=2, miny=0, maxy=1),
            geometry.box(minx=2, maxx=3, miny=0, maxy=1),
            geometry.box(minx=0, maxx=1, miny=1, maxy=2),
            geometry.box(minx=1, maxx=2, miny=1, maxy=2),
            geometry.box(minx=2, maxx=3, miny=1, maxy=2),
            geometry.box(minx=0, maxx=1, miny=2, maxy=3),
            geometry.box(minx=1, maxx=2, miny=2, maxy=3),
            geometry.box(minx=2, maxx=3, miny=2, maxy=3),
        ],
        index=["SW", "S", "SE", "W", "CENTER", "E", "NW", "N", "NE"],  # compass directions
        crs=WGS84_CRS,
    )
    return regions


@pytest.fixture  # type: ignore
def rounded_regions_fixture() -> gpd.GeoDataFrame:
    """
    Get 9 small rounded square regions.

    This GeoDataFrame represents 9 small squares with buffer on a cartesian plane in a 3 by 3 grid
    pattern. Regions are adjacent by edges, but not by vertices. Regions are given compass
    directions. Visually it looks like this ("C" means "CENTER"):

    [["NW", "N", "NE"],
     [ "W", "C",  "E"],
     ["SW", "S", "SE"]]

    Returns:
        GeoDataFrame: A GeoDataFrame containing rounded regions.
    """
    regions = gpd.GeoDataFrame(
        geometry=[
            geometry.box(minx=0, maxx=0.5, miny=0, maxy=0.5).buffer(0.25),
            geometry.box(minx=1, maxx=1.5, miny=0, maxy=0.5).buffer(0.25),
            geometry.box(minx=2, maxx=2.5, miny=0, maxy=0.5).buffer(0.25),
            geometry.box(minx=0, maxx=0.5, miny=1, maxy=1.5).buffer(0.25),
            geometry.box(minx=1, maxx=1.5, miny=1, maxy=1.5).buffer(0.25),
            geometry.box(minx=2, maxx=2.5, miny=1, maxy=1.5).buffer(0.25),
            geometry.box(minx=0, maxx=0.5, miny=2, maxy=2.5).buffer(0.25),
            geometry.box(minx=1, maxx=1.5, miny=2, maxy=2.5).buffer(0.25),
            geometry.box(minx=2, maxx=2.5, miny=2, maxy=2.5).buffer(0.25),
        ],
        index=["SW", "S", "SE", "W", "CENTER", "E", "NW", "N", "NE"],  # compass directions
        crs=WGS84_CRS,
    )
    return regions


def test_no_geometry_gdf_attribute_error(no_geometry_gdf: gpd.GeoDataFrame) -> None:
    """Test checks if GeoDataFrames without geometry are disallowed."""
    with pytest.raises(ValueError):
        AdjacencyNeighbourhood(no_geometry_gdf)


def test_empty_gdf_empty_set(empty_gdf: gpd.GeoDataFrame) -> None:
    """Test checks if empty GeoDataFrames return empty neighbourhoods."""
    neighbourhood = AdjacencyNeighbourhood(empty_gdf)
    assert not neighbourhood.get_neighbours(1)


def test_empty_gdf_empty_set_include_center(empty_gdf: gpd.GeoDataFrame) -> None:
    """Test checks if empty GeoDataFrames return empty neighbourhoods."""
    neighbourhood = AdjacencyNeighbourhood(empty_gdf, include_center=True)
    assert not neighbourhood.get_neighbours(1)


def test_lazy_loading_empty_set(squares_regions_fixture: gpd.GeoDataFrame) -> None:
    """Test checks if lookup table is empty after init."""
    neighbourhood = AdjacencyNeighbourhood(squares_regions_fixture)
    assert not neighbourhood.lookup


def test_lazy_loading_empty_set_include_center(squares_regions_fixture: gpd.GeoDataFrame) -> None:
    """Test checks if lookup table is empty after init."""
    neighbourhood = AdjacencyNeighbourhood(squares_regions_fixture, include_center=True)
    assert not neighbourhood.lookup


def test_adjacency_lazy_loading(rounded_regions_fixture: gpd.GeoDataFrame) -> None:
    """Test checks if lookup table is lazily populated."""
    neighbourhood = AdjacencyNeighbourhood(rounded_regions_fixture)
    neighbours = neighbourhood.get_neighbours("SW")
    assert neighbours == {"W", "S"}
    assert neighbourhood.lookup == {
        "SW": {"W", "S"},
    }


def test_adjacency_lazy_loading_include_center(rounded_regions_fixture: gpd.GeoDataFrame) -> None:
    """Test checks if lookup table is lazily populated."""
    neighbourhood = AdjacencyNeighbourhood(rounded_regions_fixture, include_center=True)
    neighbours = neighbourhood.get_neighbours("SW")
    assert neighbours == {"W", "S", "SW"}
    assert neighbourhood.lookup == {
        "SW": {"W", "S", "SW"},
    }


def test_generate_all_neighbourhoods_rounded_regions(
    rounded_regions_fixture: gpd.GeoDataFrame,
) -> None:
    """Test checks `generate_neighbourhoods` function with rounded regions."""
    neighbourhood = AdjacencyNeighbourhood(rounded_regions_fixture)
    neighbourhood.generate_neighbourhoods()
    assert neighbourhood.lookup == {
        "SW": {"W", "S"},
        "S": {"SW", "CENTER", "SE"},
        "SE": {"E", "S"},
        "W": {"SW", "CENTER", "NW"},
        "CENTER": {"N", "W", "E", "S"},
        "E": {"CENTER", "NE", "SE"},
        "NW": {"N", "W"},
        "N": {"CENTER", "NE", "NW"},
        "NE": {"N", "E"},
    }


def test_generate_all_neighbourhoods_squares_regions(
    squares_regions_fixture: gpd.GeoDataFrame,
) -> None:
    """Test checks `generate_neighbourhoods` function with square regions."""
    neighbourhood = AdjacencyNeighbourhood(squares_regions_fixture)
    neighbourhood.generate_neighbourhoods()
    assert neighbourhood.lookup == {
        "SW": {"W", "S", "CENTER"},
        "S": {"SW", "W", "CENTER", "SE", "E"},
        "SE": {"E", "S", "CENTER"},
        "W": {"N", "SW", "S", "CENTER", "NW"},
        "CENTER": {"SW", "N", "W", "S", "SE", "E", "NW", "NE"},
        "E": {"N", "S", "CENTER", "SE", "NE"},
        "NW": {"W", "N", "CENTER"},
        "N": {"W", "CENTER", "E", "NW", "NE"},
        "NE": {"E", "N", "CENTER"},
    }


@pytest.mark.parametrize(  # type: ignore
    "index,distance,expected,expected_include_center",
    [
        ("SW", -2, set(), set()),
        ("SW", -1, set(), set()),
        ("SW", 0, set(), {"SW"}),
        ("SW", 1, {"S", "W"}, {"S", "W"}),
        ("SW", 2, {"CENTER", "SE", "NW"}, {"CENTER", "SE", "NW"}),
        ("SW", 3, {"N", "E"}, {"N", "E"}),
        ("SW", 4, {"NE"}, {"NE"}),
        ("SW", 5, set(), set()),
        ("CENTER", 0, set(), {"CENTER"}),
        ("CENTER", 1, {"N", "E", "S", "W"}, {"N", "E", "S", "W"}),
        ("CENTER", 2, {"NW", "NE", "SW", "SE"}, {"NW", "NE", "SW", "SE"}),
        ("CENTER", 3, set(), set()),
        ("N", 0, set(), {"N"}),
        ("N", 1, {"NW", "NE", "CENTER"}, {"NW", "NE", "CENTER"}),
        ("N", 2, {"E", "S", "W"}, {"E", "S", "W"}),
        ("N", 3, {"SE", "SW"}, {"SE", "SW"}),
        ("N", 4, set(), set()),
    ],
)
def test_adjacency_lazy_loading_at_distance(
    index: str,
    distance: int,
    expected: set[str],
    expected_include_center: set[str],
    rounded_regions_fixture: gpd.GeoDataFrame,
) -> None:
    """Test checks `get_neighbours_at_distance` function with rounded regions."""
    neighbourhood = AdjacencyNeighbourhood(rounded_regions_fixture)
    neighbours = neighbourhood.get_neighbours_at_distance(index, distance)
    assert neighbours == expected
    neighbours = neighbourhood.get_neighbours_at_distance(index, distance, include_center=True)
    assert neighbours == expected_include_center

    neighbourhood = AdjacencyNeighbourhood(rounded_regions_fixture, include_center=True)
    neighbours = neighbourhood.get_neighbours_at_distance(index, distance)
    assert neighbours == expected_include_center
    neighbours = neighbourhood.get_neighbours_at_distance(index, distance, include_center=False)
    assert neighbours == expected


@pytest.mark.parametrize(  # type: ignore
    "index,distance,expected,expected_include_center",
    [
        ("SW", -2, set(), set()),
        ("SW", -1, set(), set()),
        ("SW", 0, set(), {"SW"}),
        ("SW", 1, {"S", "W"}, {"SW", "S", "W"}),
        ("SW", 2, {"S", "W", "CENTER", "SE", "NW"}, {"SW", "S", "W", "CENTER", "SE", "NW"}),
        (
            "SW",
            3,
            {"S", "W", "CENTER", "SE", "NW", "N", "E"},
            {"SW", "S", "W", "CENTER", "SE", "NW", "N", "E"},
        ),
        (
            "SW",
            4,
            {"S", "W", "CENTER", "SE", "NW", "N", "E", "NE"},
            {"SW", "S", "W", "CENTER", "SE", "NW", "N", "E", "NE"},
        ),
        (
            "SW",
            5,
            {"S", "W", "CENTER", "SE", "NW", "N", "E", "NE"},
            {"SW", "S", "W", "CENTER", "SE", "NW", "N", "E", "NE"},
        ),
        ("CENTER", 0, set(), {"CENTER"}),
        ("CENTER", 1, {"N", "E", "S", "W"}, {"CENTER", "N", "E", "S", "W"}),
        (
            "CENTER",
            2,
            {"N", "E", "S", "W", "NW", "NE", "SW", "SE"},
            {"CENTER", "N", "E", "S", "W", "NW", "NE", "SW", "SE"},
        ),
        (
            "CENTER",
            3,
            {"N", "E", "S", "W", "NW", "NE", "SW", "SE"},
            {"CENTER", "N", "E", "S", "W", "NW", "NE", "SW", "SE"},
        ),
        ("N", 0, set(), {"N"}),
        ("N", 1, {"NW", "NE", "CENTER"}, {"N", "NW", "NE", "CENTER"}),
        ("N", 2, {"NW", "NE", "CENTER", "E", "S", "W"}, {"N", "NW", "NE", "CENTER", "E", "S", "W"}),
        (
            "N",
            3,
            {"NW", "NE", "CENTER", "E", "S", "W", "SE", "SW"},
            {"N", "NW", "NE", "CENTER", "E", "S", "W", "SE", "SW"},
        ),
        (
            "N",
            4,
            {"NW", "NE", "CENTER", "E", "S", "W", "SE", "SW"},
            {"N", "NW", "NE", "CENTER", "E", "S", "W", "SE", "SW"},
        ),
    ],
)
def test_adjacency_lazy_loading_up_to_distance(
    index: str,
    distance: int,
    expected: set[str],
    expected_include_center: set[str],
    rounded_regions_fixture: gpd.GeoDataFrame,
) -> None:
    """Test checks `get_neighbours_up_to_distance` function with rounded regions."""
    neighbourhood = AdjacencyNeighbourhood(rounded_regions_fixture)
    neighbours = neighbourhood.get_neighbours_up_to_distance(index, distance)
    assert neighbours == expected
    neighbours = neighbourhood.get_neighbours_up_to_distance(index, distance, include_center=True)
    assert neighbours == expected_include_center

    neighbourhood = AdjacencyNeighbourhood(rounded_regions_fixture, include_center=True)
    neighbours = neighbourhood.get_neighbours_up_to_distance(index, distance)
    assert neighbours == expected_include_center
    neighbours = neighbourhood.get_neighbours_up_to_distance(index, distance, include_center=False)
    assert neighbours == expected



"""GeoVex HexagonalDataset tests."""

from contextlib import nullcontext as does_not_raise
from typing import Any

import geopandas as gpd
import pandas as pd
import pytest

from srai.embedders.geovex.dataset import HexagonalDataset
from srai.h3 import get_local_ij_index
from srai.neighbourhoods import AdjacencyNeighbourhood, H3Neighbourhood

ROOT_REGION = "891e205194bffff"
RING_DISTANCE = 25


@pytest.fixture  # type: ignore
def empty_gdf() -> gpd.GeoDataFrame:
    """Get empty GeoDataFrame."""
    return gpd.GeoDataFrame(geometry=[])


@pytest.fixture  # type: ignore
def regions_data_df() -> pd.DataFrame:
    """
    Get example regions for testing.

    Expects first region to be a ROOT_REGION.
    """
    neighbourhood = H3Neighbourhood()
    regions_indices = [ROOT_REGION]
    regions_indices.extend(
        list(neighbourhood.get_neighbours_up_to_distance(ROOT_REGION, RING_DISTANCE))
    )
    data_df = pd.DataFrame(
        list(range(len(regions_indices))), index=regions_indices, columns=["data"]
    )
    return data_df


@pytest.mark.parametrize(  # type: ignore
    "ring_distance,expectation",
    [
        (-1, pytest.raises(ValueError)),
        (0, pytest.raises(ValueError)),
        (1, pytest.raises(ValueError)),
        (2, does_not_raise()),
        (3, does_not_raise()),
    ],
)
def test_raises_with_incorrect_ring_distance(
    ring_distance: int,
    expectation: Any,
) -> None:
    """Test if HexagonalDataset checks ring_distance correctness."""
    data = pd.DataFrame()
    neighbourhood = H3Neighbourhood()
    with expectation:
        HexagonalDataset(data, neighbourhood, neighbor_k_ring=ring_distance)


@pytest.mark.parametrize(
    "ring_distance",
    [2, 3, 4],
)  # type: ignore
def test_dataset_length(ring_distance: int, regions_data_df: pd.DataFrame):
    """Test if HexagonalDataset constructs lookup tables correctly."""
    neighbourhood: H3Neighbourhood = H3Neighbourhood(regions_data_df)
    dataset = HexagonalDataset(regions_data_df, neighbourhood, neighbor_k_ring=ring_distance)  # type: ignore
    assert len(dataset) == len(
        neighbourhood.get_neighbours_up_to_distance(
            ROOT_REGION, distance=RING_DISTANCE - ring_distance, include_center=True, unchecked=True
        )
    )


@pytest.mark.parametrize(
    "neighborhood_cls,expectation",
    [(H3Neighbourhood, does_not_raise()), (AdjacencyNeighbourhood, pytest.raises(ValueError))],
)  # type: ignore
def test_neighborhood_type(
    neighborhood_cls: Any,
    expectation: Any,
    regions_data_df: pd.DataFrame,
    empty_gdf: gpd.GeoDataFrame,
) -> None:
    """Test if HexagonalDataset correctly accepts only H3Neighborhoods."""
    neighborhood = neighborhood_cls(empty_gdf)
    with expectation:
        HexagonalDataset(regions_data_df, neighborhood)


def test_dataset_item(regions_data_df: pd.DataFrame) -> None:
    """Test if HexagonalDataset constructs lookup tables correctly."""
    import numpy as np

    ring_distance = 2

    neighbourhood = H3Neighbourhood(regions_data_df)
    dataset = HexagonalDataset(regions_data_df, neighbourhood, neighbor_k_ring=ring_distance)  # type: ignore
    item = next(iter(dataset)).detach().numpy()
    # flatten it out and get the corresponding hexagons
    cells = regions_data_df.reset_index().set_index("data").loc[item.reshape(-1).tolist()].values

    # it starts with the root region
    # for each of the h3s, calculate the ij index
    ijs = np.array([get_local_ij_index(ROOT_REGION, _cell) for _cell in cells])

    ijs = ijs.reshape(ring_distance * 2 + 2, ring_distance * 2 + 2, 2)

    # commpare to the transposed image in the paper
    # specifically fig. 3
    # the bottom and right are padded by 0s for even #
    desired = np.array(
        [
            [(0, 0), (0, 0), (0, 2), (1, 2), (2, 2), (0, 0)],
            [(0, 0), (-1, 1), (0, 1), (1, 1), (2, 1), (0, 0)],
            [(-2, 0), (-1, 0), (0, 0), (1, 0), (2, 0), (0, 0)],
            [(-2, -1), (-1, -1), (0, -1), (1, -1), (0, 0), (0, 0)],
            [(-2, -2), (-1, -2), (0, -2), (0, 0), (0, 0), (0, 0)],
            [(0, 0), (0, 0), (0, 0), (0, 0), (0, 0), (0, 0)],
        ]
    )
    assert np.all(ijs.transpose(1, 0, -1) == desired)



"""NeighbourDataset tests."""

from contextlib import nullcontext as does_not_raise
from typing import TYPE_CHECKING, Any

import pandas as pd
import pytest

from srai.embedders.hex2vec.neighbour_dataset import NeighbourDataset
from srai.neighbourhoods import H3Neighbourhood

if TYPE_CHECKING:  # pragma: no cover
    from pytest_mock import MockerFixture


@pytest.fixture  # type: ignore
def regions_data_df() -> pd.DataFrame:
    """Get example regions for testing."""
    root_region = "891e205194bffff"
    neighbourhood = H3Neighbourhood()
    regions_indices = list(neighbourhood.get_neighbours_up_to_distance(root_region, 25))
    regions_indices.append(root_region)
    data_df = pd.DataFrame(0, index=regions_indices, columns=["data"])
    return data_df


@pytest.mark.parametrize(  # type: ignore
    "negative_sample_k_distance,expectation",
    [
        (-1, pytest.raises(ValueError)),
        (0, pytest.raises(ValueError)),
        (1, pytest.raises(ValueError)),
        (2, does_not_raise()),
        (3, does_not_raise()),
    ],
)
def test_raises_with_incorrect_sample_k_distance(
    negative_sample_k_distance: int, expectation: Any, mocker: "MockerFixture"
) -> None:
    """Test if NeighbourDataset checks negative_sample_k_distance correctness."""
    data = pd.DataFrame()
    neighbourhood = mocker.Mock()
    with expectation:
        NeighbourDataset(data, neighbourhood, negative_sample_k_distance=negative_sample_k_distance)


@pytest.mark.parametrize("negative_sample_k_distance", [2, 3, 4])  # type: ignore
def test_lookup_tables_construction(negative_sample_k_distance: int, regions_data_df: pd.DataFrame):
    """Test if NeighbourDataset constructs lookup tables correctly."""
    neighbourhood = H3Neighbourhood(regions_data_df)
    dataset = NeighbourDataset(
        regions_data_df, neighbourhood, negative_sample_k_distance=negative_sample_k_distance
    )

    positives_correct = [
        regions_data_df.index[positive_df_loc]
        in neighbourhood.get_neighbours(regions_data_df.index[anchor_df_loc])
        for anchor_df_loc, positive_df_loc in zip(
            dataset._anchor_df_locs_lookup, dataset._positive_df_locs_lookup
        )
    ]

    assert all(positives_correct)

    excluded_correct = []
    for i, index in enumerate(regions_data_df.index):
        excluded_df_locs = dataset._excluded_from_negatives[i]
        excluded_df_indices = set(regions_data_df.index[list(excluded_df_locs)])
        excluded_correct.append(
            neighbourhood.get_neighbours_up_to_distance(index, negative_sample_k_distance)
            == excluded_df_indices
        )

    assert all(excluded_correct)


def test_dataset_length(regions_data_df: pd.DataFrame) -> None:
    """Test if NeighbourDataset has correct length."""
    neighbourhood = H3Neighbourhood(regions_data_df)
    expected_length = sum(
        len(neighbourhood.get_neighbours(region)) for region in regions_data_df.index
    )
    dataset = NeighbourDataset(regions_data_df, neighbourhood)
    assert len(dataset) == expected_length



"""
OSM Online Loader.

This module contains loader capable of loading OpenStreetMap features from Overpass.
"""

from collections.abc import Iterable
from itertools import product
from typing import Union

import geopandas as gpd
import pandas as pd
from functional import seq
from packaging import version
from shapely.geometry.base import BaseGeometry
from tqdm import tqdm

from srai._optional import import_optional_dependencies
from srai.constants import FEATURES_INDEX, GEOMETRY_COLUMN, WGS84_CRS
from srai.loaders.osm_loaders._base import OSMLoader
from srai.loaders.osm_loaders.filters import (
    GroupedOsmTagsFilter,
    OsmTagsFilter,
    merge_osm_tags_filter,
)


class OSMOnlineLoader(OSMLoader):
    """
    OSMOnlineLoader.

    OSM(OpenStreetMap)[1] online loader is a loader capable of downloading objects
    from a given area from OSM. It filters features based on OSM tags[2] in form of
    key:value pairs, that are used by OSM users to give meaning to geometries.

    This loader is a wrapper around the `osmnx` library. It uses `osmnx.geometries_from_polygon`
    to make individual queries.

    References:
        1. https://www.openstreetmap.org/
        2. https://wiki.openstreetmap.org/wiki/Tags
    """

    _PBAR_FORMAT = "Downloading {}: {}"
    _ELEMENT_TYPE_INDEX_NAME = "element_type"
    _OSMID_INDEX_NAME = "osmid"
    _RESULT_INDEX_NAMES = [_ELEMENT_TYPE_INDEX_NAME, _OSMID_INDEX_NAME]

    def __init__(self) -> None:
        """Initialize OSMOnlineLoader."""
        import_optional_dependencies(dependency_group="osm", modules=["osmnx"])

    def load(
        self,
        area: Union[BaseGeometry, Iterable[BaseGeometry], gpd.GeoSeries, gpd.GeoDataFrame],
        tags: Union[OsmTagsFilter, GroupedOsmTagsFilter],
    ) -> gpd.GeoDataFrame:
        """
        Download OSM features with specified tags for a given area.

        The loader first downloads all objects with `tags`. It returns a GeoDataFrame containing
        the `geometry` column and columns for tag keys.

        Note: Some key/value pairs might be missing from the resulting GeoDataFrame,
            simply because there are no such objects in the given area.

        Args:
            area (Union[BaseGeometry, Iterable[BaseGeometry], gpd.GeoSeries, gpd.GeoDataFrame]):
                Area for which to download objects.
            tags (Union[OsmTagsFilter, GroupedOsmTagsFilter]): A dictionary
                specifying which tags to download.
                The keys should be OSM tags (e.g. `building`, `amenity`).
                The values should either be `True` for retrieving all objects with the tag,
                string for retrieving a single tag-value pair
                or list of strings for retrieving all values specified in the list.
                `tags={'leisure': 'park}` would return parks from the area.
                `tags={'leisure': 'park, 'amenity': True, 'shop': ['bakery', 'bicycle']}`
                would return parks, all amenity types, bakeries and bicycle shops.

        Returns:
            gpd.GeoDataFrame: Downloaded features as a GeoDataFrame.
        """
        import osmnx as ox

        area_wgs84 = self._prepare_area_gdf(area)

        merged_tags = merge_osm_tags_filter(tags)

        _tags = self._flatten_tags(merged_tags)

        total_tags_num = len(_tags)
        total_queries = len(area_wgs84) * total_tags_num

        key_value_name_max_len = self._get_max_key_value_name_len(_tags)
        desc_max_len = key_value_name_max_len + len(self._PBAR_FORMAT.format("", ""))

        results = []

        osmnx_new_api = version.parse(ox.__version__) >= version.parse("1.5.0")
        osmnx_download_function = (
            ox.features_from_polygon if osmnx_new_api else ox.geometries_from_polygon
        )

        pbar = tqdm(product(area_wgs84[GEOMETRY_COLUMN], _tags), total=total_queries)
        for polygon, (key, value) in pbar:
            pbar.set_description(self._get_pbar_desc(key, value, desc_max_len))
            geometries = osmnx_download_function(polygon, {key: value})
            if not geometries.empty:
                results.append(geometries[[GEOMETRY_COLUMN, key]])

        result_gdf = self._group_gdfs(results).set_crs(WGS84_CRS)
        result_gdf = self._flatten_index(result_gdf)

        return self._parse_features_gdf_to_groups(result_gdf, tags)

    def _flatten_tags(self, tags: OsmTagsFilter) -> list[tuple[str, Union[str, bool]]]:
        tags_flat: list[tuple[str, Union[str, bool]]] = (
            seq(tags.items())
            .starmap(lambda k, v: product([k], v if isinstance(v, list) else [v]))
            .flatten()
            .list()
        )
        return tags_flat

    def _get_max_key_value_name_len(self, tags: list[tuple[str, Union[str, bool]]]) -> int:
        max_key_val_name_len: int = seq(tags).starmap(lambda k, v: len(k + str(v))).max()
        return max_key_val_name_len

    def _get_pbar_desc(self, key: str, val: Union[str, bool], max_desc_len: int) -> str:
        return self._PBAR_FORMAT.format(key, val).ljust(max_desc_len)

    def _group_gdfs(self, gdfs: list[gpd.GeoDataFrame]) -> gpd.GeoDataFrame:
        if not gdfs:
            return self._get_empty_result()
        elif len(gdfs) == 1:
            gdf = gdfs[0]
        else:
            gdf = pd.concat(gdfs)

        return gdf.groupby(self._RESULT_INDEX_NAMES).first()

    def _get_empty_result(self) -> gpd.GeoDataFrame:
        result_index = pd.MultiIndex.from_arrays(arrays=[[], []], names=self._RESULT_INDEX_NAMES)
        return gpd.GeoDataFrame(index=result_index, crs=WGS84_CRS, geometry=[])

    def _flatten_index(self, gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
        gdf = gdf.reset_index()
        gdf[FEATURES_INDEX] = (
            gdf[self._RESULT_INDEX_NAMES]
            .apply(lambda idx: "/".join(map(str, idx)), axis=1)
            .astype(str)
        )
        return gdf.set_index(FEATURES_INDEX).drop(columns=self._RESULT_INDEX_NAMES)



"""Constants for OSMWayLoader."""

import numpy as np

MPH_TO_KMH = 1.609344
INCHES_TO_METERS = 0.0254
FEET_TO_METERS = 0.3048

METADATA_COLUMNS = ["from", "to", "key", "osmid", "name", "reversed", "length", "ref"]

OSMNX_TIMEOUT: int = 180
"""The timeout interval in seconds for HTTP requests."""

OSMNX_WAY_KEYS: list[str] = [
    "bridge",
    "tunnel",
    "oneway",
    "lanes",
    "ref",
    "name",
    "highway",
    "maxspeed",
    "service",
    "access",
    "area",
    "landuse",
    "width",
    "est_width",
    "junction",
    # missing in the original OSMnx config
    "surface",
    "footway",
    "bicycle",
    "lit",
]
"""Extended OSMnx way tags."""


OSM_WAY_TAGS: dict[str, list[str]] = {
    "oneway": ["False", "True"],
    "lanes": list(map(str, range(1, 21))),
    "highway": [
        "motorway",
        "motorway_link",
        "trunk",
        "trunk_link",
        "primary",
        "primary_link",
        "secondary",
        "secondary_link",
        "tertiary",
        "tertiary_link",
        "residential",
        "unclassified",
        "living_street",
        "service",
        "pedestrian",
        "track",
        "bus_guideway",
        "escape",
        "raceway",
        "road",
        "busway",
        "footway",
        "bridleway",
        "steps",
        "corridor",
        "path",
    ],
    "maxspeed": [
        "5",
        "7",
        "10",
        "15",
        "20",
        "30",
        "40",
        "50",
        "60",
        "70",
        "80",
        "90",
        "100",
        "110",
        "120",
        "130",
        "140",
    ],
    "bridge": [
        "yes",
        "aqueduct",
        "boardwalk",
        "cantilever",
        "covered",
        "low_water_crossing",
        "movable",
        "trestle",
        "viaduct",
        "simple_brunnel",
    ],
    "access": [
        "yes",
        "no",
        "private",
        "public",
        "permissive",
        "permit",
        "destination",
        "delivery",
        "customers",
        "designated",
        "use_sidepath",
        "dismount",
        "agricultural",
        "forestry",
        "discouraged",
        "unknown",
        "restricted",
        "official",
        "students",
        "licence",
        "license",
        "exclusion_zone",
        "dismount",
    ],
    "junction": [
        "yes",
        "roundabout",
        "circular",
        "spui",
        "jughandle",
        "filter",
        "intersection",
        "crossing",
        "uncontrolled",
    ],
    "width": list(np.arange(0, 30.5, 0.5).astype("str")),
    "tunnel": ["building_passage", "yes", "avalanche_protector"],
    "surface": [
        "paved",
        "asphalt",
        "chipseal",
        "concrete",
        "concrete:lanes",
        "concrete:plates",
        "paving_stones",
        "sett",
        "unhewn_cobblestone",
        "cobblestone",
        "metal",
        "wood",
        "stepping_stones",
        "rubber",
        "unpaved",
        "compacted",
        "fine_gravel",
        "gravel",
        "rock",
        "pebblestone",
        "ground",
        "dirt",
        "earth",
        "grass",
        "grass_paver",
        "mud",
        "sand",
        "woodchips",
        "snow",
        "ice",
        "salt",
    ],
    "bicycle": [
        "yes",
        "no",
        "designated",
        "use_sidepath",
        "dismount",
        "permissive",
        "private",
        "customers",
        "destination",
        "official",
    ],
    "lit": ["yes", "no", "sunset-sunrise", "24/7", "automatic", "disused", "limited", "interval"],
}
"""
Acceptable features assembled based on [1] and [2].

Assembled values that are officially defined in the wiki and are related to the `way` OSM type.

References:
    1. https://taginfo.openstreetmap.org/
    2. https://wiki.openstreetmap.org/wiki/Main_Page
"""


OSM_IMPLICIT_MAXSPEEDS = {
    "AR:urban": "40",
    "AR:rural": "110",
    "AR:urban:primary": "60",
    "AR:urban:secondary": "60",
    "AT:urban": "50",
    "AT:rural": "100",
    "AT:bicycle_road": "30",
    "AT:trunk": "100",
    "AT:motorway": "130",
    "AT:zone:30": "30",
    "AT:walk": "5",
    "walk": "5",
    "AU:urban": "None",
    "AU:rural": "None",
    "BE:living_street": "20",
    "BE:cyclestreet": "30",
    "BE:zone30": "30",
    "BE:trunk": "120",
    "BE:motorway": "120",
    "BG:urban": "50",
    "BG:rural": "90",
    "BG:living_street": "20",
    "BG:trunk": "120",
    "BG:motorway": "140",
    "BY:urban": "60",
    "BY:rural": "90",
    "BY:living_street": "20",
    "BY:motorway": "110",
    "CH:rural": "80",
    "CH:trunk": "100",
    "CH:motorway": "120",
    "CZ:urban": "50",
    "CZ:rural": "90",
    "CZ:pedestrian_zone": "20",
    "CZ:living_street": "20",
    "CZ:urban_motorway": "80",
    "CZ:urban_trunk": "80",
    "CZ:trunk": "110",
    "CZ:motorway": "130",
    "DE:urban": "50",
    "DE:rural": "100",
    "DE:living_street": "7",
    "DE:bicycle_road": "30",
    "DE:motorway": "None",
    "DK:urban": "50",
    "DK:rural": "80",
    "DK:motorway": "130",
    "EE:urban": "50",
    "EE:rural": "90",
    "ES:urban": "50",
    "ES:rural": "90",
    "ES:living_street": "20",
    "ES:zone30": "30",
    "ES:trunk": "90",
    "ES:motorway": "120",
    "FI:urban": "50",
    "FI:rural": "80",
    "FI:trunk": "100",
    "FI:motorway": "120",
    "FR:urban": "50",
    "FR:rural": "80",
    "FR:zone30": "30",
    "FR:motorway": "None",
    "GB:nsl_restricted": "30 mph",
    "GB:motorway": "70 mph",
    "GB:nsl_dual": "70 mph",
    "GB:nsl_single": "60 mph",
    "GR:urban": "50",
    "GR:rural": "90",
    "GR:trunk": "110",
    "GR:motorway": "130",
    "HU:urban": "50",
    "HU:rural": "90",
    "HU:living_street": "20",
    "HU:trunk": "110",
    "HU:motorway": "130",
    "IT:urban": "50",
    "IT:rural": "90",
    "IT:trunk": "110",
    "IT:motorway": "130",
    "JP:nsl": "60",
    "JP:express": "100",
    "LT:urban": "50",
    "LT:rural": "90",
    "NO:urban": "50",
    "NO:rural": "80",
    "PH:urban": "30",
    "PH:rural": "80",
    "PH:express": "100",
    "PL:urban": "50",
    "PT:urban": "50",
    "PT:rural": "90",
    "PT:trunk": "100",
    "PT:motorway": "120",
    "RO:urban": "50",
    "RO:rural": "90",
    "RO:trunk": "100",
    "RO:motorway": "130",
    "RS:urban": "50",
    "RS:rural": "80",
    "RS:living_street": "10",
    "RS:trunk": "100",
    "RS:motorway": "130",
    "RU:urban": "60",
    "RU:rural": "90",
    "RU:living_street": "20",
    "RU:motorway": "110",
    "SE:urban": "50",
    "SE:rural": "70",
    "SE:trunk": "90",
    "SE:motorway": "110",
    "SI:urban": "50",
    "SI:rural": "90",
    "SI:trunk": "110",
    "SI:motorway": "130",
    "SK:urban": "50",
    "SK:rural": "90",
    "SK:living_street": "20",
    "SK:trunk": "90",
    "SK:motorway": "130",
    "SK:motorway_urban": "90",
    "UA:urban": "50",
    "UA:rural": "90",
    "UA:living_street": "20",
    "UA:trunk": "110",
    "UA:motorway": "130",
    "UK:motorway": "70 mph",
    "UK:nsl_dual": "70 mph",
    "UK:nsl_single": "60 mph",
    "UZ:urban": "70",
    "UZ:rural": "100",
    "UZ:living_street": "30",
    "UZ:motorway": "110",
}
"""
Mapping for implicit maxspeeds.

The mapping is incomplete (see Notes section). Mostly taken from [1].

Notes:
    Consider using better (full) table in machine readable format available in [2].

References:
    1. https://wiki.openstreetmap.org/wiki/Key:maxspeed#Implicit_maxspeed_values
    2. https://wiki.openstreetmap.org/wiki/Default_speed_limits
"""



"""
Intersection Joiner.

This module contains intersection joiner implementation.
"""

import geopandas as gpd
import pandas as pd

from srai.constants import FEATURES_INDEX, GEOMETRY_COLUMN, REGIONS_INDEX
from srai.joiners import Joiner


class IntersectionJoiner(Joiner):
    """
    Intersection Joiner.

    Intersection Joiner allows to join two GeoDataFrames and find all overlapping geometries. It
    does not apply any grouping or aggregation.
    """

    def transform(
        self, regions: gpd.GeoDataFrame, features: gpd.GeoDataFrame, return_geom: bool = False
    ) -> gpd.GeoDataFrame:
        """
        Join features to regions based on an 'intersects' predicate.

        Does not apply any grouping to regions.

        Args:
            regions (gpd.GeoDataFrame): regions with which features are joined
            features (gpd.GeoDataFrame): features to be joined
            return_geom (bool): whether to return geometry of the joined features.
                Defaults to False.

        Returns:
            GeoDataFrame with an intersection of regions and features, which contains
            a MultiIndex and optionaly a geometry with the intersection
        """
        if GEOMETRY_COLUMN not in regions.columns:
            raise ValueError("Regions must have a geometry column.")
        if GEOMETRY_COLUMN not in features.columns:
            raise ValueError("Features must have a geometry column.")

        if len(regions) == 0:
            raise ValueError("Regions must not be empty.")
        if len(features) == 0:
            raise ValueError("Features must not be empty.")

        result_gdf: gpd.GeoDataFrame

        if return_geom:
            result_gdf = self._join_with_geom(regions, features)
        else:
            result_gdf = self._join_without_geom(regions, features)

        return result_gdf

    def _join_with_geom(
        self, regions: gpd.GeoDataFrame, features: gpd.GeoDataFrame
    ) -> gpd.GeoDataFrame:
        """
        Join features to regions with returning an intersecting geometry.

        Args:
            regions (gpd.GeoDataFrame): regions with which features are joined
            features (gpd.GeoDataFrame): features to be joined

        Returns:
            GeoDataFrame with an intersection of regions and features, which contains
            a MultiIndex and a geometry with the intersection
        """
        joined_parts = [
            gpd.overlay(
                single[[GEOMETRY_COLUMN]].reset_index(names=FEATURES_INDEX),
                regions[[GEOMETRY_COLUMN]].reset_index(names=REGIONS_INDEX),
                how="intersection",
                keep_geom_type=False,
            ).set_index([REGIONS_INDEX, FEATURES_INDEX])
            for _, single in features.groupby(features[GEOMETRY_COLUMN].geom_type)
        ]

        joint = gpd.GeoDataFrame(pd.concat(joined_parts, ignore_index=False))
        return joint

    def _join_without_geom(
        self, regions: gpd.GeoDataFrame, features: gpd.GeoDataFrame
    ) -> gpd.GeoDataFrame:
        """
        Join features to regions without intersection caclulation.

        Args:
            regions (gpd.GeoDataFrame): regions with which features are joined
            features (gpd.GeoDataFrame): features to be joined

        Returns:
            GeoDataFrame with an intersection of regions and features, which contains
            a MultiIndex
        """
        joint = (
            gpd.sjoin(
                regions.reset_index(names=REGIONS_INDEX),
                features[[GEOMETRY_COLUMN]].reset_index(names=FEATURES_INDEX),
                how="inner",
                predicate="intersects",
            )
            .set_index([REGIONS_INDEX, FEATURES_INDEX])
            .drop(columns=["index_right", GEOMETRY_COLUMN])
        )
        return joint
